-- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-- Written by Kimberly L. Tripp - all rights reserved.
-- 
-- For more scripts and sample code, check out 
-- 	http://www.SQLSkills.com
-- For "More than Just Training," see the Industry Experts at 
-- 	http://www.SolidQualityLearning.com
--
-- Disclaimer - Thoroughly test this script, execute at your own risk.
-- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

/* 
In this exercise you will walk through a scenario where a relatively low percentage
of rows are modified and the fragmentation created as a result is high.
*/
USE Credit
go

--1) Create a working version of the member table called member2. 

IF OBJECTPROPERTY(object_id('member2'), 'IsUserTable') = 1
	DROP TABLE dbo.member2 
go
CREATE TABLE dbo.member2 (
	member_no 	numeric_id 		NOT NULL ,
	lastname 		shortstring 	NOT NULL ,
	firstname 		shortstring 	NOT NULL ,
	middleinitial 	letter 			NULL ,
	street 			shortstring 	NOT NULL ,
	city 			shortstring 	NOT NULL ,
	state_prov 		statecode 		NOT NULL ,
	country 		countrycode 	NOT NULL ,
	mail_code 		mailcode 		NOT NULL ,
	phone_no 		phonenumber 	NULL ,
	issue_dt 		datetime 		NOT NULL ,
	expr_dt 		datetime 		NOT NULL ,
	region_no 		numeric_id 		NOT NULL ,
	corp_no 		numeric_id 		NULL ,
	prev_balance 	money 			NULL ,
	curr_balance 	money 			NULL ,
	member_code 	status_code 	NOT NULL 
)
GO

--2) Populate the member2 table with the same data from member.

INSERT member2
	SELECT * FROM member
go

--3) Create indexes on the member2 table to simulate a realworld environment.

ALTER TABLE member2
ADD CONSTRAINT member2PK
	PRIMARY KEY CLUSTERED (member_no)
go

CREATE INDEX member2NameInd 
ON member2(lastname, firstname, middleinitial)
go

CREATE INDEX member2RegionFK
ON member2(region_no)
go

CREATE INDEX member2CorpFK
ON member2(corp_no)
go

-- Verify the indexes
exec sp_helpindex member2
go
select object_name(id) as tablename, indid, name
from sysindexes
where object_id('member2') = id

--4) Verify the fragmentation of the indexes

-- Prior to SQL Server 2000
DECLARE @ObjID	int
SELECT @ObjID = object_id('member2')
DBCC SHOWCONTIG(@ObjID)
--or
-- New Syntax in SQL Server 2000
DBCC SHOWCONTIG('member2') --WITH TABLERESULTS (new FINALLY in ss2k)

	/* NOTE: there should not be any fragmentation currently as the indexes 
	have just been built.
        */

-- 5) Simulate modifications against the member2 table by performing a single update
-- statement that updates roughly 5% of the table. Also note that the script shows the 
-- total time it takes to perform the modification (just as an fyi).

DECLARE @StartTime	datetime, 
	@EndTime	datetime
SELECT @StartTime = getdate()
UPDATE member2
	SET street = '1234 Main St.',
		city = 'Anywhere'
WHERE member_no % 19 = 0
SELECT NumRowsMod = @@RowCount
SELECT @EndTime = getdate()
SELECT 'TOTAL TIME (ms)' = datediff (ms, @StartTime, @EndTime)
go

-- 6) Next let's look at the [negative] affect that this has had on our member2 table.

DBCC SHOWCONTIG('member2') --WITH TABLERESULTS (new FINALLY in ss2k)
go
-- 	NOTE: the fragmentation should be extreme as the Scan Density will probably 
-- 	show a VERY low number. 

-- 7) Should we rebuild our index? And should we set a FILLFACTOR? 

-- 	ABSOLUTELY - YES and YES! 

-- 8) So how do we choose our fillfactor?

-- Lecture?

-- 9) Now that we've decided it is important to rebuild our indexes we need to look at 
-- options. There are multiple ways to rebuild an index, all of which are shown below with
-- syntax and notes on why you might use one over the other. For this lab we are going to 
-- use the most common option - DBCC DBREINDEX.

-- 	* DROP and RECREATE
-- 	    NOTES - This is more challenging than it sounds considering our clustered index is our Primary Key. 
-- 		We cannot drop this Primary Key IF there are foreign keys which reference it unless we drop all of the
-- 		foreign keys first. Since this is NOT realistic SQL Server has DBCC DBREINDEX and CREATE with 
-- 		DROP_EXISTING. Additionally because the non-clustered indexes depend on the clustered 
--		index value dropping and recreating will cause the non-clustered indexes to be rebuilt TWICE!
-- 
-- 	* DBCC DBREINDEX (added in SQL Server 6.5) 
--		DBCC DBREINDEX allows you to rebuild an index easily. The syntax
--		is the easiest to automate and the rebuild is identical to that of 
--		CREATE with DROP_EXISTING. However, if you need to change the
--		definition of the clustered index then you must use CREATE with 
--		DROP_EXISTING.
-- 	    NOTES:
--		* If the Clustering Key is NOT unique then ALL non-clustered indexes
--		will be rebuilt when the clustered index is rebuilt (regardless of which
-- 		command you use).
--		* The best way to run DBCC DBREINDEX is to run it without
-- 		a fillfactor and in fact the indexes will be rebuilt using their Original 
--		Fill Factor. This "OrigFillFactor" is the same fillfactor that was used 
-- 		the last time the index was built. This gives you the greatest flexibility.

-- 	    syntax
		DBCC DBREINDEX (member2, member2PK, 85) -- because we've never used a FF for this index prior to now
-- 		DBCC DBREINDEX (member2, member2PK) -- Use the OrigFillFactor
--
-- 	* CREATE with DROP_EXISTING (new in SQL Server 7.0)
-- 	    NOTES - This was added to minimize the recreation of non-clustered indexes when only the leaf level of the
-- 		clustered was fragmented. In 2000sp2 - this is the same as DBCC DBREINDEX. 
--		Pre-2000 sp2 it was a bug that all of the non-clustered indexes were rebuilt.

--        CREATE WITH DROP_EXISTING when you DO NOT CHANGE the definition of the CL index
--            WILL NOT rebuild the nonclustered indexes UNLESS you are rebuilding the CL index on a non-unique key
--        CREATE WITH DROP_EXISTING when you CHANGE the definition of the CL index
--            WILL rebuild the nonclustered indexes using their OrigFillFactor (from sysindexes)

-- 	    syntax 
-- 		CREATE UNIQUE CLUSTERED INDEX member2pk
-- 			ON member2(member_no)
-- 			WITH FILLFACTOR = 85, DROP_EXISTING
-- 
--	IF original index is 
-- 		CREATE CLUSTERED INDEX member2CL
-- 			ON member2(lastname, firstname, middleinitial)
--
-- 		CREATE UNIQUE CLUSTERED INDEX member2CL  --requires the same name
-- 			ON member2(member_no)
-- 			WITH FILLFACTOR = 85, DROP_EXISTING
--
--	IF you want to change the name of the index then check out the syntax for
-- 	sp_rename. It's not doc'ed extremely well for renaming indexes so make
--	sure you use the correct syntax - oldname is in the form of table.index
--	and new name just requires the name of the index (not table.).
--
-- 	The best time to perform ANY of these modifications is when the system is at it lowest usage. All of the above options
--	place an EXCLUSIVE Lock on the base table (unless you're building a nonclustered index and then in that case it's a 
--	SHARED lock on the base table). BUT wait...there's a 2000 option!
--
--	If the system is a 24x7 system then SQL Server 2000 offers a new option called DBCC INDEXDEFRAG. 
--
-- 	* DBCC INDEXDEFRAG (new in SQL Server 2000)
-- 	    NOTES - This option, when a table is heavily fragmented, will take longer to defrag a table than the CREATE
-- 		with DROP_EXISTING however it will allow access to the table while it is being rebuilt. This option relies on an 
--		already set FillFactor...so you must rebuild the index once OR you must have created the index with the fillfactor set.
--		CREATE TABLE does support FILLFACTOR settings when creating the clustered and nonclustered indexes through
--		constraints. TRICK - You can also modify the system table to make this change---change the sysindexes table for the
--		OrigFillFactor column. Make sure you verify your SQL update by writing it as a select first!
-- 
--   	    syntax
--            DBCC INDEXDEFRAG(credit, member2, member2cl)

-- 10) Once rebuilt, check the fragmentation again.

DBCC SHOWCONTIG('member2')

-- 	NOTE: the internal fragmentation is what we expect - 90%. As that's what we chose for 
-- 	our fillfactor.

-- 11) Let's try the modification again with even more rows.

DECLARE @StartTime	datetime, 
	@EndTime	datetime
SELECT @StartTime = getdate()
UPDATE member2
	SET street = '1234 Main St.',
		city = 'Anywhere'
WHERE member_no % 17 = 0
SELECT NumRowsMod = @@RowCount
SELECT @EndTime = getdate()
SELECT 'TOTAL TIME (ms)' = datediff (ms, @StartTime, @EndTime)
go

-- You should have been able to modify more rows in a shorter period of time. 

--12) And again, check the fragmentation to see if our chosen fillfactor was appropriate.

DBCC SHOWCONTIG('member2')

--	NOTE: This modification did not create fragmentation. You can go back to step 11 and modify where % 11 = 0 
--	and check the fragmentation again... See how many modifcations you get before you begin fragmenting your 
--	data again! The main idea is that you've made free space available to the modifcations and deferred fragmentation
-- 	HOPEFULLY until the next rebuild.

-- 13) Another way to see information about indexes from sysindexes

SELECT CASE
		WHEN indid = 0 THEN 'Heap'
		WHEN indid = 1 THEN 'CL'
		ELSE 'NC ' + convert(varchar(5), indid)
	END AS IndexID, 
	object_name(id) AS ObjectName,
	name AS IndexName,
	OrigFillFactor AS 'FillFactor'
FROM sysindexes
    WHERE indid <= 250
ORDER BY 2,1

-- HAVE FUN!!!!!
-- Kimberly